/*-- $Id$ -- 

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================
 
    Copyright (C) 1999 The Apache Software Foundation. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:
 
 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
 
 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.
 
 4. The names "Fop" and  "Apache Software Foundation"  must not be used to
    endorse  or promote  products derived  from this  software without  prior
    written permission. For written permission, please contact
    apache@apache.org.
 
 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 James Tauber <jtauber@jtauber.com>. For more  information on the Apache 
 Software Foundation, please see <http://www.apache.org/>.
 
 */
package org.apache.fop.fo;

import java.util.Hashtable;
import org.apache.fop.messaging.MessageHandler;

import org.apache.fop.apps.FOPException;

public class PropertyList extends Hashtable {
  private PropertyListBuilder builder;
  private PropertyList parentPropertyList = null;
  String namespace = "";
  String element = "";

  public PropertyList(PropertyList parentPropertyList, String space, String el) {
    this.parentPropertyList = parentPropertyList;
    this.namespace = space;
    this.element = el;
  }

  /**
   * Return the value explicitly specified on this FO.
   * @param propertyName The name of the property whose value is desired.
   * @return The value if the property is explicitly set, otherwise null.
   */
  public Property getExplicit(String propertyName) {
    return (Property)super.get(propertyName);
  }

  /**
   * Return the value of this property inherited by this FO.
   * Implements the inherited-property-value function.
   * The property must be inheritable!
   * @param propertyName The name of the property whose value is desired.
   * @return The inherited value, otherwise null.
   */
  public Property getInherited(String propertyName) {
    if (builder != null) {
      if (parentPropertyList != null &&
	  builder.isInherited(namespace, element, propertyName)) {
	return parentPropertyList.get(propertyName);
      }
      else {
	// return the "initial" value
	try {
	  return builder.makeProperty(this, namespace, element, propertyName);
	} catch (org.apache.fop.apps.FOPException e) {
	  MessageHandler.errorln("Exception in getInherited(): property=" +
				 propertyName + " : " + e);
	}
      }
    }
    return null; // No builder or exception in makeProperty!
  }

  /**
   * Return the property on the current FlowObject. If it isn't set explicitly,
   * this will try to compute it based on other properties, or if it is
   * inheritable, to return the inherited value. If all else fails, it returns
   * the default value.
   */
  public Property get(String propertyName) {

    if (builder == null)
      MessageHandler.errorln("OH OH, builder has not been set");

    /* Handle request for one part of a compound property */
    int sepchar = propertyName.indexOf('.');
    String subpropName = null;
    if (sepchar > -1) {
      subpropName = propertyName.substring(sepchar+1);
      propertyName = propertyName.substring(0,sepchar);
    }

    Property p = getExplicit(propertyName);

    if (p == null) { // if not explicit
      p = this.builder.computeProperty(this,namespace, element, propertyName);
      if (p == null) { // else inherit (if has parent and is inheritable)
        if ((this.parentPropertyList != null) &&
	    (this.builder.isInherited(namespace, element, propertyName))) {
          p = this.parentPropertyList.get(propertyName); // retrieve parent's value
        } else { // default value
          try {
            p = this.builder.makeProperty(this,namespace, element,propertyName);
          } catch (FOPException e) {
            // don't know what to do here
          }
        }
      }
    }
    if (subpropName != null) {
      return this.builder.getSubpropValue(namespace, element, propertyName,
					   p, subpropName);
    }
    else return p;
  }

  public void setBuilder(PropertyListBuilder builder) {
    this.builder = builder;
  }

  public String getNameSpace()
  {
    return namespace;
  }

  public String getElement()
  {
    return element;
  }

  /**
   * Return the "nearest" specified value for the given property.
   * Implements the from-nearest-specified-value function.
   * @param propertyName The name of the property whose value is desired.
   * @return The computed value if the property is explicitly set on some
   * ancestor of the current FO, else the initial value.
   */
  public Property getNearestSpecified(String propertyName) {
    Property p = null;
    for (PropertyList plist = this; p == null && plist != null;
	   plist = plist.parentPropertyList) {
	p = plist.getExplicit(propertyName);
    }
    if (p == null) {
      // If no explicit setting found, return initial (default) value.
      try {
	p = this.builder.makeProperty(this,namespace, element,propertyName);
      } catch (FOPException e) {
	MessageHandler.errorln("Exception in getNearestSpecified(): property=" +
			       propertyName + " : " + e);
      }
    }
    return p;
  }

  /**
   * Return the value of this property on the parent of this FO.
   * Implements the from-parent function.
   * @param propertyName The name of the property whose value is desired.
   * @return The computed value on the parent or the initial value if this
   * FO is the root or is in a different namespace from its parent.
   */
  public Property getFromParent(String propertyName) {
    if (parentPropertyList != null) {
      return parentPropertyList.get(propertyName);
    }
    else if (builder != null) {
      // return the "initial" value
      try {
	return builder.makeProperty(this, namespace, element, propertyName);
      } catch (org.apache.fop.apps.FOPException e) {
	MessageHandler.errorln("Exception in getFromParent(): property=" +
			       propertyName + " : " + e);
      }
    }
    return null; // No builder or exception in makeProperty!
  }

}

